import abc
from typing import *
from typing import IO, BinaryIO
from io import UnsupportedOperation
import logging


class ScpiTransportBase(IO):
    # attribute default values
    transport_name: str = None
    transport_info: str = None
    transport_type: str = None

    encoding: str = 'iso-8859-15'
    newline: str = '\r\n'

    io: BinaryIO
    logger: logging.Logger

    @classmethod
    @abc.abstractmethod
    def discover(cls, **kwargs):
        raise NotImplementedError("Please Implement this method")

    @abc.abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__()

    def read(self, n: int = -1) -> AnyStr:
        self.logger.debug(f'read: block with size of {n} bytes')
        return self.io.read(n)

    def write(self, s: AnyStr) -> int:
        n = self.io.write(s)
        self.io.flush()
        self.logger.debug(f'write: block with size of {n} bytes')
        return n

    def readline(self, size: int = -1) -> str:
        data = self.io.readline(size)
        self.logger.debug(f'read line {str(data)[1:]}')
        return data.decode(self.encoding).rstrip(self.newline)

    def writeline(self, line: str) -> None:
        data = (line + self.newline).encode(self.encoding)
        self.logger.debug(f'write line {str(data)[1:]}')
        self.io.writelines([data])
        self.io.flush()

    def readlines(self, hint: int = ...) -> List[AnyStr]:
        raise UnsupportedOperation

    def writelines(self, lines: Iterable[AnyStr]) -> None:
        raise UnsupportedOperation

    def close(self) -> None:
        self.io.close()

    @property
    def closed(self) -> bool:
        return self.io.closed

    @property
    def mode(self) -> str:
        return self.io.mode

    @property
    def name(self) -> str:
        return self.io.name

    def isatty(self) -> bool:
        return self.io.isatty()

    def fileno(self) -> int:
        return self.io.fileno()

    def flush(self) -> None:
        return self.io.flush()

    def readable(self) -> bool:
        return self.io.readable()

    def seek(self, offset: int, whence: int = 0) -> int:
        return self.io.seek(offset, whence)

    def seekable(self) -> bool:
        return self.io.seekable()

    def tell(self) -> int:
        return self.io.tell()

    def truncate(self, size: Optional[int] = ...) -> int:
        return self.io.truncate(size)

    def writable(self) -> bool:
        return self.io.writable()

    def __next__(self) -> AnyStr:
        return self.io.__next__()

    def __iter__(self) -> Iterator[AnyStr]:
        return self.io.__iter__()

    def __enter__(self) -> IO[AnyStr]:
        return self.io.__enter__()

    def __exit__(self, t: Optional[Type[BaseException]], value: Optional[BaseException],
                 traceback) -> Optional[bool]:
        return self.io.__exit__(t, value, traceback)


# TypeVar for type hinting subclasses of ScpiTransportBase
ScpiTransports = TypeVar('ScpiTransports', bound=ScpiTransportBase)



import socket
import logging
from urllib.parse import urlparse
from typing import *
from .scpi_transport_base import ScpiTransportBase
from .scpi_resource import ScpiResource
from .scpi_exceptions import ScpiTransportException

logger = logging.getLogger(__name__)


class ScpiTcpIpTransport(ScpiTransportBase):
    DEFAULT_PORT: int = 5025

    transport_name = 'ScpiTcpIpTransport'
    transport_info = 'TCP/IP SCPI Transport'
    transport_type = 'TCP/IP'

    socket: socket.socket

    @classmethod
    def discover(cls,
                 dnssd_services: List[str] = ('_scpi-raw._tcp.local.',),
                 dnssd_timeout: float = 0.5) -> List[ScpiResource]:
        from zeroconf import Zeroconf, ServiceListener, ServiceBrowser
        import time

        class Listener(ServiceListener):
            def __init__(self):
                self.services = {}

            def remove_service(self, zeroconf, zc_type, zc_name):
                self.services.pop(zc_name)

            def add_service(self, zeroconf, zc_type, zc_name):
                self.services.update({zc_name: zeroconf.get_service_info(zc_type, zc_name)})

            def update_service(self, zeroconf, zc_type, zc_name):
                self.services.update({zc_name: zeroconf.get_service_info(zc_type, zc_name)})

        # Find devices via zeroconf mDNS
        # TODO: Implement DNS-SD for non-.local domains
        listener = Listener()
        ServiceBrowser(Zeroconf(), dnssd_services, listener=listener)

        # Wait for some time to get answers
        time.sleep(dnssd_timeout)

        return [ScpiResource(transport=ScpiTcpIpTransport,
                             location=f'dnssd:{service.name}',
                             address=f'{service.parsed_addresses()[0]}:{service.port}',
                             name=service.get_name(),
                             manufacturer=service.properties[b'Manufacturer'].decode(
                                 'utf-8') if b'Manufacturer' in service.properties else None,
                             model=service.properties[b'Model'].decode(
                                 'utf-8') if b'Model' in service.properties else None,
                             serialnum=service.properties[b'SerialNumber'].decode(
                                 'utf-8') if b'SerialNumber' in service.properties else None,
                             info=service
                             ) for service in listener.services.values()]

    def __init__(self, address: str, timeout: float = 5, **kwargs):
        super().__init__(**kwargs)
        self.logger = logging.getLogger(__name__)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.socket.settimeout(timeout)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        # Get hostname and port
        addr = urlparse('//' + address)
        if (addr.hostname is None):
            raise ScpiTransportException(f'{address} is not a valid hostname')

        try:
            # Try to connect
            self.socket.connect((addr.hostname,
                                 addr.port if addr.port is not None else self.DEFAULT_PORT))

            # Make a file interface for easier handling
            self.io = self.socket.makefile('rwb', buffering=0)
            self.io.flush()
        except OSError as e:
            raise ScpiTransportException(e) from e



from typing import NamedTuple, Any, Optional, Type, List
from .scpi_transport_base import ScpiTransports, ScpiTransportBase


class ScpiResource(NamedTuple):
    transport: Type[ScpiTransports]
    address: str
    location: Optional[str] = None
    name: Optional[str] = None
    manufacturer: Optional[str] = None
    model: Optional[str] = None
    serialnum: Optional[str] = None
    info: Optional[Any] = None

    @classmethod
    def discover(cls, **kwargs) -> List['ScpiResource']:
        """Find SCPI devices connected to system.

        By default lists all available possible SCPI devices. However behaviour can be controlled using
        kwargs supported by the transport classes."""

        transport_classes = ScpiTransportBase.__subclasses__()
        resources = []

        for transport_class in transport_classes:
            # filter keyword arguments accepted by transport and call transport's find_devices method
            args = {k:v for k, v in kwargs.items() if k in transport_class.discover.__code__.co_varnames}
            resources += transport_class.discover(**args)

        return resources

    @property
    def resource_name(self) -> str:
        return 'ScpiResource'

    @property
    def resource_type(self) -> str:
        return 'SCPI'

    @property
    def resource_info(self) -> str:
        return 'SCPI Resource'
